@@ -3,7 +3,7 @@
 #ifdef _MSC_VER
 #define _USE_MATH_DEFINES
 #endif
-#include <math.h>
+#include <cmath>
 #include "lensfun.h"
 
 typedef struct {
@@ -16,11 +16,10 @@ typedef struct {
 void mod_setup(lfFixture *lfFix, gconstpointer data)
 {
     lfFix->lens              = new lfLens();
-    lfFix->lens->CropFactor  = 1.0f;
-    lfFix->lens->AspectRatio = 1.0f;
     lfFix->lens->Type        = LF_RECTILINEAR;
 
-    lfLensCalibDistortion lensCalibDist = {LF_DIST_MODEL_POLY3, 12.0f, {0.1,0.5,0.5}};
+    lfLensCalibAttributes   lensSetting = { 0.0, 0.0, 1.0, 1.0 };
+    lfLensCalibDistortion lensCalibDist = {LF_DIST_MODEL_POLY3, 12.0f, 10.8f, false, {0.1}, lensSetting};
     lfFix->lens->AddCalibDistortion(&lensCalibDist);
 
     // width and height have to be odd, so we have a non fractional center position
@@ -54,11 +53,9 @@ void test_mod_projection_center(lfFixture* lfFix, gconstpointer data)
             if(g_test_verbose())
                 g_print("  ~ Conversion from %s -> %s \n", geom_names[j], geom_names[i]);
 
-            lfFix->mod = new lfModifier (lfFix->lens, 1.0f, lfFix->img_width, lfFix->img_height);
-            lfFix->mod->Initialize (
-                lfFix->lens, LF_PF_U8, 12.0f,
-                6.7f, 2.0f, 1.0f, geom_types[i],
-                LF_MODIFY_GEOMETRY, false);
+            lfFix->mod = new lfModifier (1.0f, lfFix->img_width, lfFix->img_height, LF_PF_U8, false);
+
+            lfFix->mod->EnableProjectionTransform(lfFix->lens, 12.0f, geom_types[i]);
 
             // check if center is not influenced
             in[0] = (lfFix->img_width-1)/2;
@@ -95,11 +92,9 @@ void test_mod_projection_borders(lfFixture* lfFix, gconstpointer data)
             if(g_test_verbose())
                 g_print("  ~ Conversion from %s -> %s \n", geom_names[j], geom_names[i]);
 
-            lfFix->mod = new lfModifier (lfFix->lens, 1.0f, lfFix->img_width, lfFix->img_height);
-            lfFix->mod->Initialize (
-                lfFix->lens, LF_PF_U8, 12.0f,
-                6.7f, 2.0f, 1.0f, geom_types[i],
-                LF_MODIFY_GEOMETRY, false);
+            lfFix->mod = new lfModifier (1.0f, lfFix->img_width, lfFix->img_height, LF_PF_U8, false);
+
+            lfFix->mod->EnableProjectionTransform(lfFix->lens, 12.0f, geom_types[i]);
 
             if (lfFix->mod->ApplyGeometryDistortion(0,0,1,1,res)) {
                 g_assert_false(std::isnan(res[0]));
