@@ -43,7 +43,6 @@ void mod_setup(lfFixture *lfFix, gconstpointer data)
   lfTestParams *p = (lfTestParams *)data;
 
   lfFix->lens             = new lfLens();
-  lfFix->lens->CropFactor = 1.0f;
   lfFix->lens->Type       = LF_RECTILINEAR;
 
   lfFix->lens->AddCalibTCA(&p->calib);
@@ -51,12 +50,9 @@ void mod_setup(lfFixture *lfFix, gconstpointer data)
   lfFix->img_height = 300;
   lfFix->img_width  = 300;
 
-  lfFix->mod = new lfModifier(lfFix->lens, 1.0f, lfFix->img_width, lfFix->img_height);
+  lfFix->mod = new lfModifier(1.0f, lfFix->img_width, lfFix->img_height, LF_PF_F32, p->reverse);
 
-  lfFix->mod->Initialize(
-    lfFix->lens, LF_PF_F32,
-    24.0f, 2.8f, 1000.0f, 1.0f, LF_RECTILINEAR,
-    LF_MODIFY_TCA, p->reverse);
+  lfFix->mod->EnableTCACorrection(lfFix->lens, 24.0f);
 
   lfFix->coordBuff = NULL;
 
@@ -160,16 +156,18 @@ int main(int argc, char **argv)
 
   for(std::vector<bool>::iterator it_reverse = reverse.begin(); it_reverse != reverse.end(); ++it_reverse)
   {
+    lfLensCalibAttributes cs = {0.0, 0.0, 1.0, 1.5};
+
     std::map<std::string, lfLensCalibTCA> tcaCalib;
     // ??? + Nikon AF-S DX Nikkor 35mm f/1.8G
     tcaCalib["LF_TCA_MODEL_LINEAR"] = (lfLensCalibTCA)
     {
-      LF_TCA_MODEL_LINEAR, 35.0f, {1.0003, 1.0003}
+      LF_TCA_MODEL_LINEAR, 35.0f, {1.0003, 1.0003}, cs
     };
     // Canon EOS 5D Mark III + Canon EF 24-70mm f/2.8L II USM
     tcaCalib["LF_TCA_MODEL_POLY3"] = (lfLensCalibTCA)
     {
-      LF_TCA_MODEL_POLY3,  24.0f, {1.0002104, 1.0000529, 0.0f, 0.0f, -0.0000220, -0.0000000}
+      LF_TCA_MODEL_POLY3,  24.0f, {1.0002104, 1.0000529, 0.0f, 0.0f, -0.0000220, -0.0000000}, cs
     };
 
     for(std::map<std::string, lfLensCalibTCA>::iterator it_tcaCalib = tcaCalib.begin(); it_tcaCalib != tcaCalib.end(); ++it_tcaCalib)
