@@ -6,12 +6,13 @@ must be converted to Lensfun C++ code of course.  However, its structure is
 already very similar to how Lensfun is structured.
 
 Usage:
-    python3 perspective_control.py <image files>
+    python3 perspective_control.py <JSON file>
 
-Every image file must have a sidecar JSON file with the extension .json with
-the following content:
+If <JSON file> is a dash "-", the JSON data is read from stdin.  The JSON
+file/data has the following structure:
 
 [
+    "image.jpeg",
     18,
     1.534,
     6,
@@ -19,16 +20,15 @@ the following content:
     [ 188,  154,  187,  154, 188,  187]
 ]
 
-Here, 18 is the focal length in mm, 1.534 is the crop factor, 6 is a scaling
-parameter, the first array are the x values of the control points, and the
-second array are the y values.  The (x, y) values must be image pixel
-coordinates.  There are 4, 6, or 8 points allowed, see
-`initialize_perspective_correction` below.
-
+Here, "image.jpeg" is the filename of the image file, 18 is the focal length in
+mm, 1.534 is the crop factor, 6 is a scaling parameter, the first array are the
+x values of the control points, and the second array are the y values.  The (x,
+y) values must be image pixel coordinates.  There are 4, 5, 6, 7, or 8 points
+allowed, see `enable_perspective_correction` below.
 """
 
 import sys, subprocess, os, array, json, multiprocessing, tempfile
-from math import sin, cos, tan, atan, sqrt, copysign, acos, log
+from math import sin, cos, tan, atan, atan2, sqrt, copysign, acos, log, isnan
 from math import pi as π
 
 def read_ppm(input_file, read_data=True):
@@ -76,13 +76,74 @@ def write_image_file(image_data, width, height, filepath):
     image_data.byteswap()
 
 
-def intersection(x1, y1, x2, y2, x3, y3, x4, y4):
-    A = x1 * y2 - y1 * x2
-    B = x3 * y4 - y3 * x4
-    C = (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4)
+def normalize(x, y):
+    norm = sqrt(x**2 + y**2)
+    try:
+        return x / norm, y / norm
+    except ZeroDivisionError:
+        return math.copysign(float("inf"), x), math.copysign(float("inf"), y)
 
-    numerator_x = (A * (x3 - x4) - B * (x1 - x2))
-    numerator_y = (A * (y3 - y4) - B * (y1 - y2))
+def central_projection(coordinates, plane_distance):
+    """Projects the coordinates on an x-y plane with the distance
+    ``plane_distance`` from the origin.  The centre of the projection is the
+    origin.
+    """
+    stretch_factor = plane_distance / coordinates[2]
+    return coordinates[0] * stretch_factor, coordinates[1] * stretch_factor
+
+def ellipse_analysis(x, y, f_normalized):
+    import numpy
+
+    # Taken from http://math.stackexchange.com/a/767126/248694
+    M = numpy.array([[x[i]**2, x[i] * y[i], y[i]**2, x[i], y[i], 1] for i in range(5)])
+    U, σ, V_ = numpy.linalg.svd(M)
+    assert all(σ > 1e-15)
+    # Taken from http://mathworld.wolfram.com/Ellipse.html, equation (15)
+    # onwards.
+    a, b, c, d, f, g = V_[-1].reshape(-1)
+    b, d, f = b/2, d/2, f/2
+
+    _D = b**2 - a * c
+    x0 = (c * d - b * f) / _D
+    y0 = (a * f - b * d) / _D
+
+    φ = 1/2 * atan(2 * b / (a - c))
+    if a > c:
+        φ += π/2
+
+    _N = 2 * (a * f**2 + c * d**2 + g * b**2 - 2 * b * d * f - a * c * g) / _D
+    _S = sqrt((a - c)**2 + 4 * b**2)
+    _R = a + c
+    a_ = sqrt(_N / (_S - _R))
+    b_ = sqrt(_N / (- _S - _R))
+    # End taken from mathworld
+    if a_ < b_:
+        a_, b_ = b_, a_
+        φ -= π/2
+    # Normalize to -π/2..π/2 so that the vertex half-plane is top or bottom
+    # rather than e.g. left or right.
+    φ = (φ + π/2) % π - π/2
+
+    # Negative sign because vertex at top (negative y values) should be
+    # default.
+    radius_vertex = - f_normalized / sqrt((a_ / b_)**2 - 1)
+    if (x[0] - x0) * (y[1] - y0) < (x[1] - x0) * (y[0] - y0):
+        radius_vertex *= -1
+
+    return radius_vertex * sin(φ), radius_vertex * cos(φ), x0, y0
+
+
+def intersection(x, y):
+    """Returns the coordinates of the intersection of the lines defined by ``x``
+    and ``y``.  Both parameters need to be exactly 4 items long.  The first two
+    items defines the one line, the last two the other.
+    """
+    A = x[0] * y[1] - y[0] * x[1]
+    B = x[2] * y[3] - y[2] * x[3]
+    C = (x[0] - x[1]) * (y[2] - y[3]) - (y[0] - y[1]) * (x[2] - x[3])
+
+    numerator_x = (A * (x[2] - x[3]) - B * (x[0] - x[1]))
+    numerator_y = (A * (y[2] - y[3]) - B * (y[0] - y[1]))
     try:
         return numerator_x / C, numerator_y / C
     except ZeroDivisionError:
@@ -119,7 +180,7 @@ def rotate_ρ_δ_ρh(ρ, δ, ρ_h, x, y, z):
     # This matrix is: R_y(ρₕ) · Rₓ(δ) · R_y(ρ)
     A11, A12, A13 = cos(ρ) * cos(ρ_h) - sin(ρ) * cos(δ) * sin(ρ_h), sin(δ) * sin(ρ_h), sin(ρ) * cos(ρ_h) + cos(ρ) * cos(δ) * sin(ρ_h)
     A21, A22, A23 = sin(ρ) * sin(δ), cos(δ), - cos(ρ) * sin(δ)
-    A31, A32, A33 = - cos(ρ) * sin(ρ_h) -  sin(ρ) * cos(δ) * cos(ρ_h), \
+    A31, A32, A33 = - cos(ρ) * sin(ρ_h) - sin(ρ) * cos(δ) * cos(ρ_h), \
                     sin(δ) * cos(ρ_h), \
                     - sin(ρ) * sin(ρ_h) + cos(ρ) * cos(δ) * cos(ρ_h)
     return \
@@ -127,30 +188,56 @@ def rotate_ρ_δ_ρh(ρ, δ, ρ_h, x, y, z):
         A21 * x + A22 * y + A23 * z, \
         A31 * x + A32 * y + A33 * z
 
-def calculate_angles(x, y, f, normalized_in_millimeters):
-    # Calculate the center of gravity of the control points
-    if len(x) == 8:
-        center_x = sum(x) / 8
-        center_y = sum(y) / 8
+def determine_ρ_h(ρ, δ, x, y, f_normalized, center_x, center_y):
+    x_, y_, z_ = [float("nan"), float("nan")], [float("nan"), float("nan")], [float("nan"), float("nan")]
+    x_[0], y_[0], z_[0] = rotate_ρ_δ(ρ, δ, x[0], y[0], f_normalized)
+    x_[1], y_[1], z_[1] = rotate_ρ_δ(ρ, δ, x[1], y[1], f_normalized)
+    if y_[0] == y_[1]:
+        if y_[0] == 0:
+            # ρ_h is undefined (horizontal great circle is on the equator).
+            return float("nan")
+        else:
+            # The horizontal vanishing point is perfectly to the left/right, so
+            # no rotation necessary.
+            return 0
     else:
+        Δx, Δz = central_projection((x_[1] - x_[0], z_[1] - z_[0], y_[1] - y_[0]), - y_[0])
+        x_h = x_[0] + Δx
+        z_h = z_[0] + Δz
+        if z_h == 0:
+            ρ_h = 0 if x_h > 0 else π
+        else:
+            ρ_h = π / 2 - atan(x_h / z_h)
+        if rotate_ρ_δ_ρh(ρ, δ, ρ_h, center_x, center_y, f_normalized)[2] < 0:
+            # We have to move the vertex to the left instead of right
+            ρ_h -= π
+        return ρ_h
+
+def calculate_angles(x, y, f_normalized):
+    number_of_control_points = len(x)
+    # Calculate the center of gravity of the control points
+    if number_of_control_points == 6:
         center_x = sum(x[:4]) / 4
         center_y = sum(y[:4]) / 4
-
-    x_v, y_v = intersection(x[0], y[0], x[1], y[1], x[2], y[2], x[3], y[3])
-    if len(x) == 8:
-        # The problem is over-determined.  I prefer the fourth line over the
-        # focal length.  Maybe this is useful in cases where the focal length
-        # is not known.
-        x_h, y_h = intersection(x[4], y[4], x[5], y[5], x[6], y[6], x[7], y[7])
-        try:
-            f_normalized = sqrt(- x_h * x_v - y_h * y_v)
-        except ValueError:
-            # FixMe: Is really always an f given (even if it is inaccurate)?
-            f_normalized = f / normalized_in_millimeters
-        else:
-            print(f_normalized * normalized_in_millimeters, "mm")
     else:
-        f_normalized = f / normalized_in_millimeters
+        center_x = sum(x) / number_of_control_points
+        center_y = sum(y) / number_of_control_points
+
+    if number_of_control_points in [5, 7]:
+        x_v, y_v, center_x, center_y = ellipse_analysis(x[:5], y[:5], f_normalized)
+    else:
+        x_v, y_v = intersection(x[:4], y[:4])
+        if number_of_control_points == 8:
+            # The problem is over-determined.  I prefer the fourth line over the
+            # focal length.  Maybe this is useful in cases where the focal length
+            # is not known.
+            x_h, y_h = intersection(x[4:], y[4:])
+            try:
+                f_normalized = sqrt(- x_h * x_v - y_h * y_v)
+            except ValueError:
+                pass
+            # else:
+            #     print(f_normalized * normalized_in_millimeters, "mm")
 
     # Calculate vertex in polar coordinates
     ρ = atan(- x_v / f_normalized)
@@ -158,44 +245,52 @@ def calculate_angles(x, y, f, normalized_in_millimeters):
     if rotate_ρ_δ(ρ, δ, center_x, center_y, f_normalized)[2] < 0:
         # We have to move the vertex into the nadir instead of the zenith.
         δ -= π
-    a = normalize(x_v - x[0], y_v - y[0])
-    b = normalize(x_v - x[2], y_v - y[2])
-    c = (a[0] + b[0], a[1] + b[1])
-    if abs(c[0]) > abs(c[1]):
-        # The first two lines denote *horizontal* lines, the last two
-        # *verticals*.  The final rotation is by final_rotation·π/2.
-        final_rotation = copysign(1, ρ)
+
+    swapped_verticals_and_horizontals = False
+
+    if number_of_control_points in [4, 6, 8]:
+        a = normalize(x_v - x[0], y_v - y[0])
+        b = normalize(x_v - x[2], y_v - y[2])
+        c = (a[0] + b[0], a[1] + b[1])
+    elif number_of_control_points == 5:
+        c = (x_v - center_x, y_v - center_y)
     else:
-        final_rotation = 0
+        c = (x[5] - x[6], y[5] - y[6])
+    if number_of_control_points == 7:
+        x5_, y5_ = central_projection(rotate_ρ_δ(ρ, δ, x[5], y[5], f_normalized), f_normalized)
+        x6_, y6_ = central_projection(rotate_ρ_δ(ρ, δ, x[6], y[6], f_normalized), f_normalized)
+        α = - atan2(y6_ - y5_, x6_ - x5_)
+        if abs(c[0]) > abs(c[1]):
+            # Find smallest rotation into horizontal
+            α = - (α - π/2) % π - π/2
+        else:
+            # Find smallest rotation into vertical
+            α = - α % π - π/2
+    elif abs(c[0]) > abs(c[1]):
+        swapped_verticals_and_horizontals = True
+        α = copysign(π/2, ρ)
+    else:
+        α = 0
 
     # Calculate angle of intersection of horizontal great circle with equator,
     # after the vertex was moved into the zenith
-    if len(x) == 4:
-        # This results from assuming that at y = 0, there is a horizontal in
-        # the picture.  y = 0 is arbitrary, but then, so is every other value.
+    if number_of_control_points == 4:
+        if swapped_verticals_and_horizontals:
+            x_perpendicular_line = [center_x, center_x]
+            y_perpendicular_line = [center_y - 1, center_y + 1]
+        else:
+            x_perpendicular_line = [center_x - 1, center_x + 1]
+            y_perpendicular_line = [center_y, center_y]
+        ρ_h = determine_ρ_h(ρ, δ, x_perpendicular_line, y_perpendicular_line, f_normalized, center_x, center_y)
+    elif number_of_control_points in [5, 7]:
         ρ_h = 0
     else:
-        z4 = z5 = f_normalized
-        x4_, y4_, z4_ = rotate_ρ_δ(ρ, δ, x[4], y[4], z4)
-        x5_, y5_, z5_ = rotate_ρ_δ(ρ, δ, x[5], y[5], z5)
-        if y4_ == y5_:
-            # The horizontal vanishing point is perfectly to the left/right, so
-            # no rotation necessary.  ρ_h is undefined if also y4_ == 0
-            # (horizontal great circle is on the equator), but we simply set
-            # ρ_h = 0 in this case, too.
-            ρ_h = 0
-        else:
-            λ = y4_ / (y4_ - y5_)
-            x_h = x4_ + λ * (x5_ - x4_)
-            z_h = z4_ + λ * (z5_ - z4_)
-            if z_h == 0:
-                ρ_h = 0 if x_h > 0 else π
-            else:
-                ρ_h = π / 2 - atan(x_h / z_h)
-            if rotate_ρ_δ_ρh(ρ, δ, ρ_h, center_x, center_y, f_normalized)[2] < 0:
-                # We have to move the vertex to the left instead of right
-                ρ_h -= π
-    return ρ, δ, ρ_h, f_normalized, final_rotation, center_x, center_y
+        ρ_h = determine_ρ_h(ρ, δ, x[4:6], y[4:6], f_normalized, center_x, center_y)
+        if isnan(ρ_h) and number_of_control_points == 8:
+            ρ_h = determine_ρ_h(ρ, δ, x[6:8], y[6:8], f_normalized, center_x, center_y)
+    if isnan(ρ_h):
+        ρ_h = 0
+    return ρ, δ, ρ_h, α, center_x, center_y, f_normalized
 
 def generate_rotation_matrix(ρ1, δ, ρ2, d):
     """Returns a rotation matrix which combines three rotations.  First, around the
@@ -239,13 +334,6 @@ def generate_rotation_matrix(ρ1, δ, ρ2, d):
            A21, A22, A23, \
            A31, A32, A33
 
-def normalize(x, y):
-    norm = sqrt(x**2 + y**2)
-    try:
-        return x / norm, y / norm
-    except ZeroDivisionError:
-        return math.copysign(float("inf"), x), math.copysign(float("inf"), y)
-
 class Modifier:
 
     """First, the coordinate system.  It is right-handed.  The centre of the image
@@ -290,7 +378,7 @@ class Modifier:
 
     And then there is the d parameter which allows finetuning of the
     correction.  It is between -1 and 1, with 0 meaning full correction, -1 no
-    correction, and 1 increased rotation by factor 1.25.  For example, if you
+    correction, and 1 increased rotation by factor 1.24.  For example, if you
     have tilted the camera by 40°, d = -1 means the original tilt of 40°, d = 0
     means no tilt (perfect correction), and d = 1 means a tilt of 10° in the
     opposite direction (over-correction).  This way, one can finetune the slope
@@ -299,20 +387,20 @@ class Modifier:
     """
 
     def __init__(self, crop_factor, width, height):
-        assert width > height
+        self.size = min(width, height)
         self.crop_factor, self.width, self.height = crop_factor, width, height
-        aspect_ratio = width / height
+        aspect_ratio = width / height if width > height else height / width
         aspect_ratio_correction = sqrt(1 + aspect_ratio**2)
         self.normalized_in_millimeters = sqrt(36**2 + 24**2) / 2 / aspect_ratio_correction / self.crop_factor
-        self.norm_scale = 2 / (height - 1)
-        self.norm_unscale = (height - 1) / 2
-        self.center_x = width / height
-        self.center_y = 1
+        self.norm_scale = 2 / (self.size - 1)
+        self.norm_unscale = (self.size - 1) / 2
+        self.center_x = width / self.size
+        self.center_y = height / self.size
 
     def initialize(self, focal_length):
-        self.focal_length = focal_length
+        self.f_normalized = focal_length / self.normalized_in_millimeters
 
-    def initialize_perspective_correction(self, x, y, d):
+    def enable_perspective_correction(self, x, y, d):
         """Depending on the number of control points given, there are three possible
         modes:
 
@@ -326,7 +414,17 @@ class Modifier:
         horizontal line.  The focal length given is discarded and calculated from
         the control points.
 
-        If the lines constructed from the first four control points are more
+        5 control points: They all must lie on an ellipse that is actually a
+        rotated circle.  If they lie in a clockwise ordering, the vertex is
+        assumed to be above the ellipse centre in the image.  If they lie in a
+        counter-clockwise ordering, the vertex is assumed to be below the
+        ellipse.  The focal length is needed to find the proper vertex.
+
+        7 control points: c0 to c4 like above.  c5 and c6 define a horizontal
+        line which is used to rotate the final image nicely.
+
+        If the lines constructed from the first four control points for the 4,
+        6 and 8 points case (or the last two for the 7 points case) are more
         horizontal than vertical, in all of the above, "horizontal" and
         "vertical" need to be swapped.
 
@@ -347,10 +445,9 @@ class Modifier:
 
         The d parameter is supposed to be offered to the user as a slider.  It
         can take values from -1 to +1.  0 denotes the perfect correction.  -1
-        is the unchanged image.  +1 is an increase of the tilting angle by 25%.
-
+        is the unchanged image.  +1 is an increase of the tilting angle by 24%.
         """
-        if self.focal_length <= 0 or len(x) not in [4, 6, 8] or len(x) != len(y):
+        if  len(x) not in [4, 5, 6, 7, 8] or len(x) != len(y) or self.f_normalized <= 0 and len(x) != 8:
             # Don't add any callback
             return False
         if d <= - 1:
@@ -365,8 +462,8 @@ class Modifier:
         x = [value * self.norm_scale - self.center_x for value in x]
         y = [value * self.norm_scale - self.center_y for value in y]
 
-        ρ, δ, ρ_h, f_normalized, final_rotation, center_of_control_points_x, center_of_control_points_y = \
-                calculate_angles(x, y, self.focal_length, self.normalized_in_millimeters)
+        ρ, δ, ρ_h, α, center_of_control_points_x, center_of_control_points_y, f_normalized = \
+                calculate_angles(x, y, self.f_normalized)
 
         # Transform center point to get shift
         z = rotate_ρ_δ_ρh(ρ, δ, ρ_h, 0, 0, f_normalized)[2]
@@ -381,33 +478,31 @@ class Modifier:
         A31, A32, A33 = generate_rotation_matrix(ρ, δ, ρ_h, d)
 
         if new_image_center == "old image center":
-            x, y, z = A13 * f_normalized, A23 * f_normalized, A33 * f_normalized
+            center_coords = A13 * f_normalized, A23 * f_normalized, A33 * f_normalized
         elif new_image_center == "control points center":
-            x, y, z = A11 * center_of_control_points_x + A12 * center_of_control_points_y + A13 * f_normalized, \
-                      A21 * center_of_control_points_x + A22 * center_of_control_points_y + A23 * f_normalized, \
-                      A31 * center_of_control_points_x + A32 * center_of_control_points_y + A33 * f_normalized
-        if z <= 0:
+            center_coords = \
+                    A11 * center_of_control_points_x + A12 * center_of_control_points_y + A13 * f_normalized, \
+                    A21 * center_of_control_points_x + A22 * center_of_control_points_y + A23 * f_normalized, \
+                    A31 * center_of_control_points_x + A32 * center_of_control_points_y + A33 * f_normalized
+        if center_coords[2] <= 0:
             return False
         # This is the mapping scale in the image center
-        mapping_scale = f_normalized / z
-        # Central projection through the origin
-        Δa = - x * mapping_scale
-        Δb = - y * mapping_scale
+        mapping_scale = f_normalized / center_coords[2]
 
-#        print(ρ * 180 / π, δ * 180 / π, ρ_h * 180 / π, final_rotation * 90)
+#        print(ρ * 180 / π, δ * 180 / π, ρ_h * 180 / π, α * 90)
 
         # Finally, generate a rotation matrix in backward (lookup) direction
         A11, A12, A13, \
         A21, A22, A23, \
         A31, A32, A33 = generate_rotation_matrix(- ρ_h, - δ, - ρ, d)
 
-        if final_rotation:
-            # This matrix is: R_y(- ρ) · Rₓ(- δ) · R_y(- ρₕ) ·
-            # R_z(final_rotation·π/2).  final_rotation is eigher -1 or +1.
-            A11, A12, A13 = final_rotation * A12, - final_rotation * A11, A13
-            A21, A22, A23 = final_rotation * A22, - final_rotation * A21, A23
-            A31, A32, A33 = final_rotation * A32, - final_rotation * A31, A33
-            Δa, Δb = final_rotation * Δb, - final_rotation * Δa
+        # Now we append the final rotation by α.  This matrix is: R_y(- ρ) ·
+        # Rₓ(- δ) · R_y(- ρₕ) · R_z(α).
+        A11, A12, A13 = cos(α) * A11 + sin(α) * A12, - sin(α) * A11 + cos(α) * A12, A13
+        A21, A22, A23 = cos(α) * A21 + sin(α) * A22, - sin(α) * A21 + cos(α) * A22, A23
+        A31, A32, A33 = cos(α) * A31 + sin(α) * A32, - sin(α) * A31 + cos(α) * A32, A33
+        Δa, Δb = central_projection(center_coords, f_normalized)
+        Δa, Δb = cos(α) * Δa + sin(α) * Δb, - sin(α) * Δa + cos(α) * Δb
 
         # The occurances of mapping_scale here avoid an additional
         # multiplication in the inner loop of perspective_correction_callback.
@@ -425,16 +520,16 @@ class Modifier:
         f_normalized = data[9]
         Δa, Δb = data[10:12]
         for i in range(count):
-            x = iocoord[offset + i * 2] - Δa
-            y = iocoord[offset + i * 2 + 1] - Δb
+            x = iocoord[offset + i * 2] + Δa
+            y = iocoord[offset + i * 2 + 1] + Δb
             z_ = A31 * x + A32 * y + A33 * f_normalized
             if z_ > 0:
                 x_ = A11 * x + A12 * y + A13 * f_normalized
                 y_ = A21 * x + A22 * y + A23 * f_normalized
-                # Central projection through the origin
-                stretch_factor = f_normalized / z_
-                iocoord[offset + i * 2] = x_ * stretch_factor
-                iocoord[offset + i * 2 + 1] = y_ * stretch_factor
+                # The following call slows down by 5% in comparison to the
+                # inline representation, however, it should not be a penalty in
+                # C++, and it adds readability.
+                iocoord[offset + i * 2], iocoord[offset + i * 2 + 1] = central_projection((x_, y_, z_), f_normalized)
             else:
                 iocoord[offset + i * 2] = iocoord[offset + i * 2 + 1] = float("nan")
 
@@ -462,12 +557,14 @@ class Modifier:
                 offset += 2
             y += self.norm_scale
 
-def process_image(filepath, d, index):
-    image_data, width, height = read_image_file(filepath)
-    f, crop_factor, shrinking, x, y = json.load(open(os.path.splitext(filepath)[0] + ".json"))
+def process_image(json_filepath, d, index):
+    input_file = open(json_filepath) if json_filepath != "-" else sys.stdin
+    image_filepath, f, crop_factor, shrinking, x, y = json.load(input_file)
+    image_filepath = os.path.join(os.path.dirname(json_filepath), image_filepath)
+    image_data, width, height = read_image_file(image_filepath)
     modifier = Modifier(crop_factor, width, height)
     modifier.initialize(f)
-    if modifier.initialize_perspective_correction(x, y, d):
+    if modifier.enable_perspective_correction(x, y, d):
         res = array.array("f", width * height * 2 * [0])
         modifier.scaling_factor = shrinking
         modifier.apply_perspective_correction(0, 0, width, height, res)
@@ -489,8 +586,8 @@ def process_image(filepath, d, index):
                     destination_image_data[destination_offset + 2] = image_data[image_offset + 2]
     else:
         destination_image_data = image_data
-    basename, extension = os.path.splitext(filepath)
-    write_image_file(destination_image_data, width, height, "{}_{:03}_{}{}".format(basename, index, d, extension))
+    basename = os.path.splitext(json_filepath)[0]
+    write_image_file(destination_image_data, width, height, "{}_{:03}_{}.png".format(basename, index, d))
 
 pool = multiprocessing.Pool()
 results = set()
