@@ -135,9 +135,6 @@ def get_database_elements():
                                        float(calibration_element.attrib["aperture"]) == aperture and \
                                        float(calibration_element.attrib["distance"]) == distance:
                                         vignetting_element = calibration_element
-                                    elif calibration_element.tag == "real-focal-length" and \
-                                       float(calibration_element.attrib["focal"]) == focal_length:
-                                        real_focal_length_element = calibration_element
                                     elif calibration_element.tag == "field_of_view" and \
                                        float(calibration_element.attrib["focal"]) == focal_length:
                                         fov_element = calibration_element
@@ -228,11 +225,17 @@ def get_real_focal_length():
         else:
             assert lens_type == "rectilinear"
             result = half_width_in_millimeters / tan(fov / 2)
-        return result * hugin_correction
+        return result
+    elif distortion_element is not None and "real-focal" in distortion_element.attrib:
+        return get_float_attribute(distortion_element, "real-focal")
     else:
-        # Interpret the nominal focal length Hugin-compatibly; it's guesswork
-        # and a fallback anyway.
         return focal_length * hugin_correction
+real_focal_length = get_real_focal_length()
+
+# This factor converts Lensfun coordinates into real-world coordinates on the
+# camera sensor, in units of the focal length.
+normalized_in_focal_lengths = \
+    sqrt(36**2 + 24**2) / 2 / sqrt(lens_aspect_ratio**2 + 1) / lens_cropfactor / real_focal_length
 
 def get_projection_function():
     projection = None
@@ -278,6 +281,16 @@ def get_distortion_function():
             k2 = get_float_attribute(distortion_element, "k2")
             def distortion(r):
                 return r * (1 + k1 * r**2 + k2 * r**4)
+        elif model == "acm":
+            k1 = get_float_attribute(distortion_element, "k1")
+            k2 = get_float_attribute(distortion_element, "k2")
+            k3 = get_float_attribute(distortion_element, "k3")
+            k4 = get_float_attribute(distortion_element, "k4")
+            k5 = get_float_attribute(distortion_element, "k5")
+            assert k4 == 0 and k5 == 0, Exception("We do not support non-radial terms.")
+            def distortion(r):
+                r *= normalized_in_focal_lengths
+                return r * (1 + k1 * r**2 + k2 * r**4 + k3 * r**6) / normalized_in_focal_lengths
     return distortion
 distortion = get_distortion_function()
 
@@ -306,6 +319,27 @@ def get_tca_functions():
             vb = get_float_attribute(tca_element, "vb", 1)
             def tca_blue(r):
                 return r**3 * bb + r**2 * cb + r * vb
+        elif model == "acm":
+            α0 = get_float_attribute(tca_element, "alpha0", 1)
+            α1 = get_float_attribute(tca_element, "alpha1")
+            α2 = get_float_attribute(tca_element, "alpha2")
+            α3 = get_float_attribute(tca_element, "alpha3")
+            α4 = get_float_attribute(tca_element, "alpha4")
+            α5 = get_float_attribute(tca_element, "alpha5")
+            assert α4 == 0 and α5 == 0, Exception("We do not support non-radial terms.")
+            def tca_red(r):
+                r *= normalized_in_focal_lengths
+                return α0 * r * (1 + α1 * r**2 + α2 * r**4 + α3 * r**6) / normalized_in_focal_lengths
+            β0 = get_float_attribute(tca_element, "beta0", 1)
+            β1 = get_float_attribute(tca_element, "beta1")
+            β2 = get_float_attribute(tca_element, "beta2")
+            β3 = get_float_attribute(tca_element, "beta3")
+            β4 = get_float_attribute(tca_element, "beta4")
+            β5 = get_float_attribute(tca_element, "beta5")
+            assert β4 == 0 and β5 == 0, Exception("We do not support non-radial terms.")
+            def tca_blue(r):
+                r *= normalized_in_focal_lengths
+                return β0 * r * (1 + β1 * r**2 + β2 * r**4 + β3 * r**6) / normalized_in_focal_lengths
     return tca_red, tca_blue
 tca_red, tca_blue = get_tca_functions()
 
@@ -320,6 +354,13 @@ def get_vignetting_function():
             k3 = get_float_attribute(vignetting_element, "k3")
             def vignetting(r):
                 return 1 + k1 * r**2 + k2 * r**4 + k3 * r**6
+        elif model == "acm":
+            α1 = get_float_attribute(vignetting_element, "alpha1")
+            α2 = get_float_attribute(vignetting_element, "alpha2")
+            α3 = get_float_attribute(vignetting_element, "alpha3")
+            def vignetting(r):
+                r *= sqrt(aspect_ratio**2 + 1) * normalized_in_focal_lengths
+                return 1 + α1 * r**2 + α2 * r**4 + α3 * r**6
     return vignetting
 vignetting = get_vignetting_function()
 
@@ -432,7 +473,7 @@ class Image:
 
     def create_grid(self, distortion, projection, tca_red, tca_blue):
         full_frame_diagonal = sqrt(36**2 + 24**2)
-        diagonal_by_focal_length = full_frame_diagonal / 2 / camera_cropfactor / (get_real_focal_length() / hugin_correction)
+        diagonal_by_focal_length = full_frame_diagonal / 2 / camera_cropfactor / real_focal_length
         def apply_lens_projection(x, y):
             r_vignetting = sqrt(x**2 + y**2) / self.aspect_ratio_correction
             ϑ = atan(r_vignetting * diagonal_by_focal_length)
@@ -449,6 +490,11 @@ class Image:
                     x, y = apply_lens_projection(x, y)
 
                 # 2. Distortion on top of that
+                # 2a. Undo the scaling in Hugin-based distortion models,
+                # so that the focal length is still correct
+                x /= hugin_correction
+                y /= hugin_correction
+                # 2b. Apply distortion model
                 r = sqrt(x**2 + y**2) * self.ar_plus_cf_correction
                 r_distorted = distortion(r)
                 distortion_scaling = r_distorted / r
